#!/usr/bin/python3

import extra_data as ex
import nx5d.h5like.xfel as xfel

import pytest

# Unit test for the XFEL data module (mostly nx5d.h5like.xfel).
# Needs XFEL data and API, predominantly extra_data.
# Tests therefore only work within the XFEL environment.
#
# Call with 'pytest -m xfel'

@pytest.fixture(scope="class")
def xfel_test_run():
    '''Which dataset to use for unit testing'''
    
    return { 'proposal': 3491,
             'run': 149,
             'data': 'all' } # proposal, run


@pytest.fixture(scope="class")
def xfel_test_data(xfel_test_run):
    '''Selection to use for unit testing (from dataset above)'''
    
    run = ex.open_run(**xfel_test_run)
    return run.select([
        ('MID_DET_AGIPD1M-1/DET/*CH0:xtdf', '*'),
        ('SA2_XTD1_XGM/XGM/DOOCS:output', '*'),
        ('MID_EXP_EPIX-1/DET/RECEIVER:daqOutput','*'),
        ('MID_EXP_SAM/MDL/DOWNSAMPLER', '*')
    ])


@pytest.mark.xfel
class TestExdatH5:
    @pytest.fixture(scope="class", autouse=True)
    def exdat_h5(self, xfel_test_data):
        '''Returns an ExdatH5 instance for testing'''
        return xfel.ExdatH5(xfel_test_data)

    @pytest.fixture(scope="class")
    def simple_dataset(self, exdat_h5):
        ''' Returns a 'simple' data set'''
        return exdat_h5["/instrument/selection/SA2_XTD1_XGM/XGM/DOOCS:output/data.xTD"]

        
    def test_xgm_available(self, exdat_h5):
        assert "instrument" in exdat_h5
        node = exdat_h5["/instrument/selection/SA2_XTD1_XGM/XGM/DOOCS:output"]
        #print (node.keys())
        
        assert "data.xTD" in node
        data = node['data.xTD']
        #print(data.shape)

        
    def test_slicing(self, simple_dataset):
        '''Tests simple slicing and arithmetics on a dataset'''
        
        subdata = simple_dataset[13:15]
        #print(subdata.shape)
        assert len(simple_dataset.shape) == 2
        assert len(subdata.shape) == 2

        nr = subdata.sum()
        nr2 = (subdata*2).sum()
        assert (nr*1.9 < nr2) and (nr*2.1 > nr2)


    def test_multimod_detect(self, exdat_h5):
        '''Discovery of multimodal detectors.
        '''

        # Step 1: Multimodal data should *not* be available at "measurement/$PATH/" anymore
        with pytest.raises(KeyError):
            regular = exdat_h5["/instrument/MID_DET_AGIPD1M-1/DET/"]

        with pytest.raises(KeyError):
            regular = exdat_h5["/instrument/selection/MID_DET_AGIPD1M-1/DET/"]

            
        # Step 2: Dedicated detector entries should be at "instrument/$DETECTOR/...",
        # and detector images linked to "measurement/$DETECTOR"
        det = exdat_h5["/instrument/multimod/MID_DET_AGIPD1M-1"]
        img1 = det['image.data']
        assert len(img1.shape) == 4 # train_pulse, module, w, h

        
    def test_multimod_slice(self, exdat_h5):
        '''Slicint on multi-module data (necessary because this
        is a different Dataset codebase). This will take a loooong time.
        '''
        img = exdat_h5["/instrument/multimod/MID_DET_AGIPD1M-1/image.data"]

        subdata = img[13:29]
        #print(subdata.shape)
        assert len(subdata.shape) == len(img.shape)
        assert subdata.shape[1:] == img.shape[1:]

        nr  =  subdata.sum()
        nr2 = (subdata*2).sum()
        assert (nr*1.9 < nr2) and (nr*2.1 > nr2)



#!/usr/bin/python3

'''
Experiment geometry defintion dictionary as needed by `LazyQMap`.
Strongly influenced by typical `xrayutilities` API. This was 
the correct geometry at one point in time (for proposal 3381/Bargheer
in April 2023), but might have changed for *your* experiment.
'''
experimentTemplate = {
    "goniometerAxes": ('y-', 'z+', 'x+'),
    "detectorTARAxes": ("y-", None, None),

    "imageAxes": ("y-", "x+"),
    "imageSize": "$/detector_size",
    "imageCenter": "$/detector_centre",

    # same unit as imageChannelSize
    "imageDistance": "@/sdd",

    # same unit as imageDistance (mm)    
    "imageChannelSize": ("@/pixel_size", "@/pixel_size"),

    "imageSize": "@/detector_size",

    'sampleFaceUp': 'x-',

    'beamDirection': (0, 0, 1),
    'sampleNormal': (1, 0, 0),

    "beamEnergy": "@/photon_energy",

    'goniometerAngles': {'omega': '@/omega', 'chi': '@/chi', 'phi': '@/phi'},
    'detectorTARAngles': {'2theta': '@/twotheta', },
}



#!/usr/bin/python3

import silx.io.commonh5 as ch5
import numpy as np

from os import path, walk
from glob import glob
from itertools import chain

import h5py
import re

import logging

''' Selected data formats for the PETRA3 synchrotron at DESY.
'''

def read_fio(f):
    ''' Reads a FIO file and returns an xarray dataset.

    Returns:
        Two dictionaries `(parameters, columns)`, where `parameters` contains all the
        `%p` section data of the FIO file, and `columns` contains arrays with all
        the `%d` section data.
    '''

    # Parameters (%p section stuff).
    # Eacy line is a key=value entry containing either a number, or an array as value.
    def param_eval(data):
        ''' Parse a %p line: return either a single float value, or an array. '''

        parts = data.split('=')
        assert len(parts) == 2
        
        key = parts[-2].strip()
        val = parts[-1].strip()

        try:
            if val[0] == '[':
                return {
                    key: np.array([
                        float(i.strip(',')) for i in val.strip('[]').split(' ')
                    ])
                }
            else:
                return {key: float(val)}
        except Exception as e:
            logging.error("Don't understand parameter '%s': %s" % (key, str(e)))
            return {key: val.encode('utf-8')}


    # Columns data (%d) stuff.
    # Data line is either a "Col ..." line, specifying ("Col", index, name, data_type)
    # or is an acutal space-separated list of items (numbers).
    data_types = { 'DOUBLE': float }   # Data type translators; so far, only DOUBLE
    data_column_types = []             # Here we'll store the types (one for each column)
    data_column_names = []             # Here we store the names; used for dict keys,
    def data_eval(data):
        dl = data.split(' ')
        if dl[0] == "Col":
            data_column_types.append(dl[3])
            data_column_names.append(dl[2])
            return None
        return { data_column_names[i]:\
                 np.array([data_types[data_column_types[i]](d)]) \
                 for i,d in enumerate(dl) }

    
    # Reading FIO file in several "states" demarked by the '%...' magic.
    # Each "state" has a parsing function associated with it.
    cur_state = None    
    read_states = { '%c': lambda l: None,
                    '%p': param_eval,
                    '%d': data_eval }


    # Here we store the final data
    data_dict = { "%p": dict(),
                  "%d": dict() }
    
    with (open(f) if isinstance(f, str) else f) as ff:
        for line in ff:
            l = line.strip()
            
            if l.startswith('!'):
                continue
            
            if l.startswith('%'):
                cur_state = l
                continue

            # Transform data according to current state (comment, params, column data)
            line_dict = read_states[cur_state](l)

            if line_dict is None:
                continue

            # Store data into data_dict; if there's already a key present, it's assumed
            # to contain an array item (typically only the case with column data) and
            # we just append the new item to that.
            for k,v in line_dict.items():
                ex = data_dict[cur_state].get(k, None)
                data_dict[cur_state][k] = v if ex is None else np.concatenate((ex, v))

    return data_dict['%p'], data_dict['%d']
            

class DictDataGroup(ch5.Group):
    ''' Encapsulates / creates a HDF5-like node from a Python dictionary.
    '''

    def __init__(self, name, parent, data, attrs=None):
        super().__init__(name, parent, attrs)
        
        for k,v in data.items():
            dset = v if isinstance(v, np.ndarray) else np.array(v)
            self.add_node(ch5.Dataset(k, parent=self, data=dset))


class FioGroup(ch5.Group):
    ''' Inserts a FIO file as a HDF5 group.
    '''

    def __init__(self, fio, parent, name="fio", attrs=None):
        super().__init__(name=name, parent=parent, attrs=attrs)

        for grp,data in zip(("parameters", "data"), read_fio(fio)):
            catGroup = DictDataGroup(parent=self, name=grp, data=data)
            self.add_node(catGroup)            



class H5pyRebaseGroup(ch5.Group):
    ''' Represents a HDF5-like group which is, in fact, an already existing h5py.Group.

    Essentially overrides all `ch5.Group` methods and properties to hand
    all calls over to the h5py node.
    '''

    def __init__(self, h5node, name, parent, attrs=None):
        super().__init__(name=name, parent=parent, attrs=attrs)
        self.h5node = h5node

    @property
    def attrs(self):
        return self.h5node.attrs

    def _is_editable(self):
        return False

    def __setitem__(self, *args):
        return self.h5node.__setitem__(*args)

    def __getitem__(self, *args):
        return self.h5node.__getitem__(*args)

    def __contains__(self, *args):
        return self.h5node.__contains__(*args)

    def __lens__(self):
        return self.h5node.__len__()

    def __iter__(self):
        return self.h5node.__iter__()

    def keys(self):
        return self.h5node.keys()

    def values(self):
        return self.h5node.values()

    def items(self):
        return self.h5node.items()

    def visit(self, *a, **kw):
        return self.h5node.visit(*a, **kw)

    def visititems(self, *a, **kw):
        return self.h5node.visititems(*a, **kw)
        
            
class FioH5(ch5.File):
    ''' Bastard HDF5 API fake for the "FIO" data format, e.g. for PETRA3/P08.

    Currently (as of: April 2023) the data format appears to be this:
    
      - A master `.fio` file appearing somewhere on the filesystem, containing
        variables with scalar data, variables with vector data, and some
        "scanning values" a.k.a. positioners. This is an ASCII file with
        a proprietary format, but not terribly complicated to parse.

      - A folder with the same base name as the `.fio` file (only without the
        `.fio` extension), on the same level as the file; it contains one
        or more subdirectories with detector names (e.g. `lambda/`)

      - Within the subdirectory, one or more `.nxs` files (Nexus/HDF5 files)
        usually containing detector data. Currently,
        the following subfolder structure has been seen in the wild:
        ```
        $ tree .
        .
        ├── m3_2507_00342
        │   └── lambda
        │       └── m3_2507_00342_00000.nxs
        ├── m3_2507_00342.fio
        ├── m3_2507_00744
        │   └── lambda
        │       ├── ct_file_00000.nxs
        │       └── m3_2507_00744_00000.nxs
        └── m3_2507_00744.fio
        ```
        We don't have exact information as to what the `ct_file_00000.nxs` is.
        Also, we don't have exact information as to what the `_00000.nxs`
        trail of the `.nxs` file is.

    Seamlessly integrating HDF5 files ("overlaying") with the parameters of the FIO
    files would be the king's version, but it is tricky for various reasons:
    
      - What do we do on ambiguous parameters? Error and bail out?

      - How do we easily iterate through all the sub-groups? (HDF5 has several
        distinct ways of retrieving subnodes: ["path"]["subpath"], ["path/subpath"],
        and several `.visit...()` functions.)

    We therefore stick to a simpler version for now and bring everything under the
    same HDF5 API hood (a.k.a. `FioH5`), but we don't overlay results. Instead, we
    give each component -- the FIO file for one, and each HDF5 file next -- its
    own node in the root namespace of `FioH5`.
    
    For convenience we're also bein opinionated about the HDF5 layout by default,
    rewarding a `./<basename>/<detector>/<basename><extra>.nxs` layout by
    auto-detecting such files, and automatically integrating them as
    `/<detector><extra>/...` in `FioH5`.

    On top of that, extra HDF5 files need to be integrated by manually specifying
    a key-value map of folder name to file path.

    Usage example:
    
    ```
    In [1]: from nx5d.h5like.petra3 import FioH5
    
    In [2]: fh = FioH5("m3_2507_00744.fio")
    
    In [3]: fh.keys()
    Out[3]: odict_keys(['fio', 'lambda_00001', 'lambda_00002', 'lambda_00000'])
    
    In [4]: fh['fio/data/om'].shape
    Out[4]: (61,)
    
    In [5]: fh['lambda_00000/entry/instrument/detector/data'].shape
    Out[5]: (61, 516, 1554)
    ```

    Usage with the nx5d `ScanReader`:
    
    ```
    In [7]: reader = ScanReader(FioH5, ("/var/home/florin/tmp/Downloads/"
                                        "P08_data/example/raw/m3_2507_00744.fio",))

    In [8]: d1 = reader.read(slice(None), omega="@/fio/data/om",
                       images="@/lambda_00000/entry/instrument/detector/data")
    ...
    ```

    (Note that the example "m3_2507_00744.fio" provided as test data with this package's
    source code does _not_ have a dataset named `data` -- it's called `"mock_data"`
    instead, and has a reduced shape of `"(61, 100, 100)"` for size reasons.)
    '''


    def __init__(self, fio, fiogrp='fio', h5extra=None):
        ''' Loads FIO (text) file, underlining its data with existing HDF5 data.

        Normally with every `.fio` file comes a directory of the same name, containing
        subdirectories named after a detector, containing HDF5 files of the same
        (or a similar) name.
        This can be overriden, or augmented with other files, using the `h5glob`
        parameter.

        Args:
        
            fio: The `.fio` file path.
        
            fiogrp: The HDF5 group name of the `.fio` file. For uniformity, we
              propose `fio`. But if this is `None`, we're using the FIO base name
              of the file (i.e. the filename only component, without extension).

            h5extra: Python `dict()` with target HDF5 folder as key, and HDF5 file
              paths as values, for HDF5 data files to integrate. If the path
              name does not begin with `/` or `.`, the data path of the current
              FIO file is prepended.      
        '''
        super().__init__()

        # base of the filename (no dirs, no extension)
        fio_base = '.'.join(path.basename(fio).split('.')[:-1])

        # base directory of where the current FIO file's HDF5 stuff is located
        fio_datadir = path.join(path.dirname(fio), fio_base)

        # Searching for <base>/<device>/<base><extra>.nxs. This will produce a
        # nested dict(): { 'device': { 'extra': <file path...> } }.
        extra_re = re.compile(fio_base+"(.*)[.]nxs")
        devices = {
            path.basename(device) : {
                extra_re.match(path.basename(f)).groups()[0] : f
                for f in glob(path.join(device, fio_base+"*.nxs"))
            } 
            for device in glob(path.join(fio_datadir, '*'))
        }

        # This is a flat {<device><extra>: <file path>, ...} map.
        flat_devices = { dev+ext:devices[dev][ext] \
                         for dev in devices \
                         for ext in devices[dev] }
        
        if h5extra is not None:
            devices.update(h5extra)

        self.fioGroup = FioGroup(fio, name=fiogrp or fio_base, parent=self)
        self.add_node(self.fioGroup)            
        
        # Open the full list of HDF5 files specified as a base.
        self.h5sub = { k:h5py.File(p, 'r') for k,p in flat_devices.items() }
        
        self.h5nodes = { }
        for h5name,h5file in self.h5sub.items():
            node = H5pyRebaseGroup(h5file["/"], name=h5name, parent=self)
            self.add_node(node)
            self.h5nodes[h5name] = node


    def _get(self, name, getlink):
        # Overrides original _get() to deal with H5pyRebaseGroup(). We essentially
        # end up here when we have a multi-path __getitem__ (e.g. self["h5file/entry"],
        # where "h5file" is the rebase group, and "entry" a node within that group.
        
        # Quick n dirty strategy is to just check the first element of "name" against
        # the self.h5sub keys, and just pass the rest on to the subnode if found

        paths = name.split('/')
        first = paths[0]
        rest = '/'.join(paths[1:])
        
        if first in self.h5nodes.keys() and len(paths) > 1:
            return self.h5nodes[first].__getitem__(rest)
        else:
            return super()._get(name, getlink)



#!/usr/bin/python3

#
#        nx5d - The NX5 Duct Tape
#        Copyright (C) 2022-2023 Florin Boariu
#
#        This program is free software: you can redistribute it and/or modify
#        it under the terms of the GNU General Public License as published by
#        the Free Software Foundation, either version 3 of the License, or
#        (at your option) any later version.
#
#        This program is distributed in the hope that it will be useful,
#        but WITHOUT ANY WARRANTY; without even the implied warranty of
#        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#        GNU General Public License for more details.
#
#        You should have received a copy of the GNU General Public License
#        along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

'''
Experiment configuration template to use with `nx5d.xrd.data.ExperimentSetup`.
Contains goniometer and detector layout for the Pilatus detector, as well
as HDF5 addresses for the goniometer and detector angles.
'''
ExperimentTemplate = {
    "goniometerAxes": ('x+', 'y+', 'z+'),
    
    "detectorTARAxes": ("x+", None, None),
    
    "imageAxes": ("x-", "z-"),
    "imageSize": (195, 487),
    "imageCenter": (90, 245),

    # This could also be used instead of 'imageChannelSize' below.
    # It's the same physical quantity, but in degrees/channel
    # instead of relative length.
    "imageChannelSpan": None,

    "imageDistance": "@{positioners}/PilatusYOffset",
    "imageChannelSize": (0.172, 0.172), # same unit as imageDistance (mm)
    
    "sampleFaceUp": 'z+',
    "beamDirection": (0, 1, 0),
    
    "sampleNormal": (0, 0, 1),
    
    "beamEnergy": "@{positioners}/monoE",

    "goniometerAngles": {
        'theta': "@{positioners}/Theta",
        'chi':   "@{positioners}/Chi",
        'phi':   "@{positioners}/Phi" },

    "detectorTARAngles": {
        'azimuth': "@{positioners}/TwoTheta",
    }
}



#!/usr/bin/python3

import xrayutilities as xu
from xarray import DataArray, Dataset
import numpy as np

import logging

'''
Generic builder to generate an "experiment setup" dictionary.

Concepts
========

The structure of an experiment setup contains several types
of data:

  - "Static" persistent device geometry data, e.g. the type
    and axes of the goniometer etc

  - "Static" transient data which is valid for a particular
    data set or scan, but not necessarily for others, e.g.
    current center of the detector plate, photon energy etc

  - "Dynamic" data which might changes from frame to frame
    (mostly positioners and auxiliary experimental data like
    ring gurrent etc.

"Static" persistent data
------------------------

All notations and conventions are analogous to `xrayutilities`
device / goniometer setup:

  - `goniometerAxes`: The direction of each of the goniometer axes in
    the `[xyz][+-]` notation. This is a variable-sized array, as there
    can be several axes in any goniometer, and `xrayutilities` apparently
    magically knows what to do.

  - `detectorTARAxes`: The direction of each of the 3 possible movement
    axes of the detector (tilt, azimuth, rotation). Note that there are
    always 3 of these axes each with a very specific purpose in
    `xrayutilties`. If your detector lacks any, complement with `None`.

  - `imageAxes`: The direction of the image axes (x and y) at zero angles.
    The positive direction of the axes should coincide with increasing pixel
    index in the data.

  - `imageCenter`: This is the positio of the center pixel, either absolute
    (integer pixel numbers), or relative to the sensor size (as specified in
    `imageAxes`). If the number is in the range 0.0..1.0, then relative
    positioning is assumed.

  - `imageChannelSpan` / `imageDistAndPixsize`: for Q transformation,
    ultimately the relation between every specific on the detectors and the angle
    of the incoming beam activating that specific pixel is needed. There
    are two distinct ways of specifying this: either using the "channel span",
    i.e. the size, in degrees, of each pixel, in horizontal/vertical direction,
    or by a distance parameter (from detector to sample) and a pixel size.
    `imageChannelSpan` is either a single number or a 2-tuple specifying
    how many degrees one channel takes. `imageChannelSize` specifies the
    spatial size of a pixel relative to the distance between the sample
    and the sensor.

  - `sampleFaceUp`: Direction of the "sample surface facing up", a.k.a.
    "sampleor" in `xrayutilities` lingo. This is the orientation of
    the sample surface at zero angles. This is either an axis notation
    (`[xyz][+-]`) or one of the special words `det`, respectively `sam`.


Example for device geometry:

```
{
  "goniometerAxes": ('y+', 'z+', 'x+'),

  "detectorTARAxes": (None, "z+", None),
  "imageAxes": ("x+", "z+"),
  "imageSize": (1024, 768),
  "imageCenter": (0.5, 0.5),
  "imageChannelSpan": (None, None),
  "imageChannelSize": (None, None),

  "sampleFaceUp": 'x+',
  "beamDirection": (0, 1, 0),
}

```

"Static" itinerant data
-----------------------

Experiment setup contains additional parameters specific to the current
experiment (as opposed to: typical for the experimentl setup). These
include:

  - `sampleNormal`: Direction perpendicular to the surface. Please also
    consult the official `xrayutilites` documentation as to the exact meaning
    of the directions.

  - `beamEnergy`: energy of the x-ray beam (in eV).

  - `detectorTARAngles`: (optional) angle values for tilt, azimuth and
    rotation of the detector. Default values for each of these angles is
    0. Only angles which are not `None` in `detectorTARAxes` are accepted.

Example for the sample geometry setup:
```
{
  "sampleNormal": (0, 0, 1),
  "beamEnergy": 15000.0,
  "detectorTARAngles": (0, 0, 0),
}
```


"Dynamic" measurement data
--------------------------

These include mostly goniometer axes, homing offsets etc. There is either
one of these per scan, or one per frame (i.e. a whole container/array of
such elements per scan). These include:

  - `goniometerAngles`: positioning for each of the angles of the
    goniometer, as specified in `goniometerAxes`

  - `detectorTARAngles`: positioning for the detector tilt, azimuth
    and rotation (as in `detectorTARAxes`)

Some of these configuration parameters (e.g. the goniometer angles) are
inherently "not fixed" as they are an essential part of the measurement.
Others, like the detector angles, may or may not be fixed (e.g. detector
might be permanently fixed at an azimuth different from 0, or may
change with every frame or with every scan, etc).

For this reason, not only scalar numbers are accepted, but actually
HDF5 dataset addresses as string. Here is an example of how such a
setup definition might look like:

```
{
  "goniometerAngles": ("Theta", "Phi", "Chi"),
  "detectorTARAngles": (0, "Psi", 0)
}
```

Here, the dataset "Theta" with respect to the default positioner path
would be used as the first goniometer angle (for ESRF-style HDF5, this
would be "instrument/positioners/Theta".)

There might also be other positioning parameters relevant for data
transformation and analysis, for instance "TwoTheta" instead of "Psi",
which would have a similar meaning except for a different
direction / offset, with respect to "Psi." In essence, it's the
responsibility of higher layers to decide what addresses / angles should
be used here.

There is no data pendant to the `imageAxes` parameter because the angle
information for each image is within the index of its own pixels.
The data set may also contain other data of interest (e.g. "SBcurrent"),
but that is viewed as something to be accessed in parallel to, and not
as a basis of, the main image data.

Each of the data fielts itself (e.g. "image", but also others like
"SBcurrent", "delay", ...) which may be imporant for the physical analysis,
but not to the transformation of data, is not a matter of `ExperimentalSetup`
(FIXME: or is it?) but one of data collection fields within
`ScanReader`.

Placeholder variables
---------------------

When using HDF5 paths as placeholders, `SetupGenerator` can use string
formatting if a dictionary of keys is supplied at `__init__`
or at `__call__`. `ScanReader` supplies the following keys, compatible
with the ESRF flavor of HDF5/Nexus data format:

  - "instrument": The top level directory containing instrument information,
    typically of HDF5 type "NXinstrument" and located at "./instrument"
    within the scan.

  - "positioners": Typically at "./instrument/positioners", usually
    important for all the angle definitions

  - "measurement": Top-level folder (of type "NXcollection") containing
    the measurement data (typically "./measurement").

Usage
=====

The idea is to make a 3-step process:

  1. Create dictionary (with keys above) that specify static
     parameters as far as possible, and contains HDF5 paths/addresses
     for the rest

  2. Pass on dictionary to `ExperimentSetup` class, which stores the
     explicitly passed values, and also offers an interface (upon
     construction or later) to override or modify keys

  3. Override `ExperimentSetup.__call__()` to export a "final" version
     of the data, hand-matched for the current scan dataset (passed on
     as parameter to `__call__()`) which does not contain anything
     else besides pure, usable values :-)

(FIXME: the whole "\\*Angles" section is actually supposd to *not* know
about loading data yet... isn't it? In a way, the Angles section is
less like "experimental setup" and more like "measurement results.")
'''        

class LazyQMap:
    '''
    Container class for a stack of experimental data frames
    and associated metadata.

    The raw / authoritative data is stored in the `Chunk.data`
    container, which has a Python `dict()` like interface.
    Beyond that, this class also offers a number of processed
    data entries that are all lazily evaluated (i.e. only processed
    on first access):

      - `qimg`: A single 2D image of the Q-space converted data stack,
        on a rectantular grid (using an `xrayutilities` gridder, meaning
        that *all* of the original image stack is being used)

      - `qxaxis`: The X-axis of the gridder that produced `qimg`

      - `qyaxis`: The Y-axis of the gridder that produced `qimg`

    All of the computed data being *lazily* evaluated means that any
    processing that must take place on the raw (i.e. untransformed)
    data can -- and must -- take place before first access to any
    of the `q...` properties. E.g. for intensity normalization, you
    could do simething like: `chunk.data['img'] *= intensity` and
    only then proceed to accessing `chunk.qimg`.
    '''

    def __init__(self, setup=None, **data):
        '''
        Parameters:
          - `data`: A dicitonary with string keys (names) and arrays
            as the data fields.

          - `setup`: Experiment definition dictionary (see for instance
            nx5d.xrd.kmc3.ExperimentTemplate).

        '''
        
        self.xdata = self.__make_dataset(**data,
                                         **(setup['goniometerAngles']),
                                         **(setup['detectorTARAngles']))

        self._data_keys = tuple([k for k in data])
        self._angle_keys = tuple([k for k in setup['goniometerAngles'].keys()] +
                                  [k for k in setup['detectorTARAngles'].keys()])
            
        self.hxrd = self.__init_experiment(setup)
        self.Ang2Q = self.hxrd.Ang2Q


    @property
    def angles(self):
        ''' dict-based access to all the "angles" fields (mimics old API).
        '''
        return { k:self.xdata[k].values for k in self._angle_keys }


    @property
    def data(self):
        ''' dict-based access to all the "data" fields (mimics old API).
        '''
        return { k:self.xdata[k].values for k in self._data_keys }
    

    def __make_dataset(self, **dsets):
        ''' Creates an `xarray.Dataset` of data sets within `dsets`.
        The first dimension of all `dsets` is required to be the same.
        This is introduced as the first dimension in the `xarray` dataset,
        with the name "__index__".
        '''
        
        tmp = next(iter(dsets.items()))
        xdata = Dataset({'index': np.array(range(tmp[1].shape[0]))})
        for k,p in dsets.items():
            data = self.__arrayify(p)
            dims = ["index"] + ["%s_%d" % (k,i) for i in range(1,len(data.shape))]
            xdata[k] = (dims, data)

        return xdata



    def __2tuple(self, data, name=""):
        ''' Returns a 2-tuple (X, Y) from a variety of data sets:
          - from a 2-tuple :-) or an array with size 2
          - from an array (N, 2)
          - from an array or tuple (2, N)
        '''
        if not hasattr(data, "__len__"):
            raise RuntimeError("%s: needs to be a 2-tuple" % name)
        
        if len(data) == 2:
            if hasattr(data[0], "__len__"):
                return ([data[0][0], data[1][0]])
            else:
                return ([data[0], data[1]])

        if len(data) > 2:
            return data[0]

        raise RuntimeError("Oops: don't know how to handle %s: %r" % (name, data))


    def __arrayify(self, val):
        return val \
            if hasattr(val, "__len__") \
            else np.array([a]*len(img))


    def __init_experiment(self, setup):
        '''
        Initializes the Experiment part (i.e. xrayutilities HXRD object etc)
        with specified device and sample geometry. The optional parameter `roi`
        restricts angle-to-Q conversion to solely this region, if it is
        specified. This is a good way to save significant amounts of computing
        time.

        '''

        logging.debug("Expriment setup: %r" % setup)

        detAxes = [x for x in filter(lambda x: x is not None, setup['detectorTARAxes'])]

        # beamEnergy is usually supposed to be a scalar, but sometimes an array
        # (one for each image) will be supplied. In that case, retrieve only the first.
        beamEnergy = setup['beamEnergy'] \
            if not hasattr(setup['beamEnergy'], "__len__") \
               else setup['beamEnergy'][0]

        qconv = xu.experiment.QConversion(sampleAxis=setup['goniometerAxes'],
                                          detectorAxis=detAxes,
                                          r_i=setup['beamDirection'],
                                          en=beamEnergy)

        hxrd = xu.HXRD(idir=setup['beamDirection'],
                       ndir=setup['sampleNormal'],
                       sampleor=setup['sampleFaceUp'],
                       qconv=qconv,
                       # Workaround for buggy xrayutilities: repeat the beam energy
                       en=beamEnergy)

        imageCenter = self.__2tuple(setup['imageCenter'], 'imageCenter')
        imageSize = self.__2tuple(setup['imageSize'], 'imageSize')
        imageDistance = setup['imageDistance'][0] \
            if hasattr(setup['imageDistance'], "__len__") \
               else setup['imageDistance']

        logging.debug("image distance %r, size %r, center at %r" % \
                      (imageDistance, imageSize, imageCenter))
        
        if imageCenter[0] <= 1 and imageCenter[1] <= 1:
            # It's a floaring-point number, relative to the detector size

            # similar considerations for imageSize as for imageCenter: expected to be
            # a 1D array with length 2, but will also accept a 2D array
            # with shape (N, 2).
            
            # FIXME: really, REALLY need to fix this. This is really ugly.
            imgs = imageSize
            assert imgs is not None
            assert imgs[0] is not None
            assert imgs[1] is not None            
            imageCenter = tuple([c*s for c,s in zip(setup['imageCenter'], imgs)])

        chSizeParm = {}
        if ('imageChannelSpan' in setup) and (setup['imageChannelSpan'] is not None):
            # channelSpan is degrees/channel, but need to pass channels/degree to Ang2Q
            imageChannelSize = self.__2tuple(setup['imageChannelSpan'],
                                             'imageChannelSpan')
            chSizeParm = {'chpdeg1': 1.0/imageChannelSpan[0],
                          'chpdeg1': 1.0/imageChannelSpan[1] }

        elif ('imageChannelSize' in setup) and (setup['imageChannelSize'] is not None):
            # Ang2Q takes one explicit distance parameter, but we're assuming that
            # channelSize is relative to the distance itself (putting the distance
            # always at 1.0 units)
            imageChannelSize = self.__2tuple(setup['imageChannelSize'],
                                             'imageChannelSize')
            logging.debug("pixel size: %r" % (imageChannelSize,))
            chSizeParm = { 'pwidth1':  imageChannelSize[0],
                           'pwidth2':  imageChannelSize[1],
                           'distance': imageDistance }

        else:
            raise RuntimeError("Experiment setup needs either the "
                               "channel span or channel size")

        tarAngles = setup.get('detectorTARAngles', [0, 0, 0])

        roi = setup.get('roi', (0, imageSize[0], 0, imageSize[1]))

        hxrd.Ang2Q.init_area(detectorDir1=setup['imageAxes'][0],
                             detectorDir2=setup['imageAxes'][1],
                             cch1=imageCenter[0],
                             cch2=imageCenter[1],
                             Nch1=imageSize[0],
                             Nch2=imageSize[1],
                             tiltazimuth=0, #tarAngles[1],
                             tilt=0,        #tarAngles[0],
                             detrot=0,      #tarAngles[2],
                             roi=roi,
                             **chSizeParm)

        return hxrd    


    def __getitem__(self, label):
        return self.xdata[label]


    def __call__(self, *datasets, **kwargs):
        return self.area_qconv(*datasets, **kwargs)

    def area_qconv(self, *datasets, **kwargs):
        '''
        Front to the ang-to-Q conversion, currently only for area data. Parameters:
        `datasets` is either empty, or a single data label. (No multiple label support
        yet.)
        
        Accepted `kwargs`:

          - `qsize`: Tuple (w, h) of the resulting Q image, or `None` (default).
            If it is `None`, the size of the original angular image(s) is used.

          - `dims`: List with dimension names for resulting Q image.
            Defaults to `["qw", "qh"]`.

          - `_gridderDict`: If specified, this is a dictionary with extra named
            parameters to be passed on to the gridder. Note that this is not portable,
            only works as long as we're using xrayutilities under the hood.

          - `_ang2qDict`: Extra set of parameters to be passed to the data-specific
            `Ang2Q` function (typically `Ang2Q.area()` for stacks of 2D datasets).
        '''

        if len(datasets) > 1:
            raise RuntimeError("Can't transform more than one dataset at a time")

        label = next(iter(datasets  if len(datasets)>0 else self._data_keys))
        img = self.xdata[label].values
        
        dims = kwargs.setdefault('dims', ('qx', 'qy', 'qz'))

        if isinstance(dims, str):
            dims = (dims,)

        if len(dims) < 1 or len(dims) > 3:
            raise RuntimeError("Bad Q-axis specification: %r" % dims)
        
        if len(img.shape) != 3:
            raise RuntimeError("Don't know how to transform objects of shape %r" \
                               % (img.shape,))


        # Make sure all angles are arrays (even those that might have been
        # passed as floats)
        ang = [self.__arrayify(self.xdata[a].values) for a in self._angle_keys]
            
        self.q = self.Ang2Q.area(*ang, **(kwargs.get('_ang2qDict', {})))

        # For transforming strings to dimension indices
        qspec = { 'x': 0, 'qx': 0,
                  'y': 1, 'qy': 1,
                  'z': 2, 'qz': 2 }

        # The list of individual Q directions/coordinates, according to user input 'dims'
        qcoord = [self.q[qspec[d]] for d in dims]

        # Default for 'qsize' is the size of the final Q-image -- make sure we account
        # for the possibly twisted order of Q-coordinates the user specified in 'dims'
        kwargs.setdefault('qsize', [img.shape[qspec[d]] for d in dims])

        # Call scheme of all the xrayutilities Gridders is pretty similar.
        Gridder = getattr(xu, "FuzzyGridder%dD" % len(dims))
        grd = Gridder(*(kwargs['qsize']))
        grd( *qcoord, img, **(kwargs.get('_gridderDict', {})) )

        # ...the tricky part is creating the DataArrays. Specifically,
        # retrieving the q coordinates from the gridder. They are in `grd`
        # attributes called 'xaxis', 'yaxis', ... according to dimension.
        # We always use qx/qy/qz for dimension keys.
        coords = {}
        for i,d in enumerate(dims):
            axname = d if len(d)>1 else 'q%s' % d
            axvals = getattr(grd, "%saxis" % ('x', 'y', 'z')[i])
            coords[axname] = axvals

        return DataArray(data=grd.data, coords=coords)

    
## Example for a class that does more than LazyQMap (namely accept angle offsets
## in its constructor), but still uses LazyQMap under the hood.
class OffsetQMap(LazyQMap):
    def __init__(self, offsets=None, **kwargs):
        super().__init__(**kwargs)
        if offsets is not None:
            for k,v in offsets.items():
                self.angles[k] += v
